weekdays = {
    "nl": {0:"ma", 1: "di", 2: "wo", 3: "do", 4: "vr", 5: "za", 6: "zo"},
    "en": {0:"Mo", 1: "Tu", 2: "We", 3: "Th", 4: "Fr", 5: "Sa", 6: "Su"},
}

holidays = {
    "nl": {
        "nieuwjaarsdag": "_new_years_day",
        "goede vrijdag": "_good_friday",
        "eerste paasdag": "_easter",
        "tweede paasdag": "_easter_monday",
        "hemelvaartsdag": "_ascension_day",
        "eerste pinksterdag": "_pentecost",
        "tweede pinksterdag": "_whit_monday",
        "eerste kerstdag": "_christmas",
        "tweede kerstdag": "_boxing_day",
        "koningsdag": "_kings_day",
        "bevrijdingsdag": "_liberation_day",
    },
    "en": {
        "new years day": "_new_years_day",
        "good friday": "_good_friday",
        "easter": "_easter",
        "easter monday": "_easter_monday",
        "ascension day": "_ascension_day",
        "pentecost": "_pentecost",
        "whit monday": "_whit_monday",
        "christmas": "_christmas",
        "boxing day": "_boxing_day",
        "kings day": "_kings_day",
        "liberationday": "_liberation_day",
    }
}